//! ## Cli definitions, arguments parsing
//! - It uses [clap](https://docs.rs/clap/latest/clap/)
//! - It includes a help options with **-h**
//! - Alows to use differents **config paths** for batch processing
//! - Alows to be used interactively via terminal as a command for a single input / output file

use clap::Parser;

use crate::{DEFAULT_CONFIG_PATH, PKG_NAME, PKG_VERSION};

/// Use [clap](https://docs.rs/clap/latest/clap/) to parse command line options with **derive mode**
#[derive(Parser, Debug)]
pub struct Cli {
    /// Config path to load targets settings (-c) command args override config values
    #[clap(short = 'c', long = "config", value_parser, display_order = 1)]
    pub config_path: Option<String>,

    /// Quiet mode only data print (-q)
    #[clap(short = 'q', long = "quiet", action, display_order = 3)]
    pub be_quiet: bool,

    /// Output path to load input data (-i)
    #[clap(short = 'i', long = "input", value_parser, display_order = 3)]
    pub input_path: Option<String>,

    /// Output path to save metric aggreates (-o)
    #[clap(short = 'o', long = "output", value_parser, display_order = 4)]
    pub output_path: Option<String>,

    /// Show version
    #[clap(short = 'v', long = "version", action, display_order = 5)]
    pub version: bool,
}
/// Collect settings for metric targets <br>
/// Only one **input** and **output** item, for more than one it is much better to use **defs::Config** for better customization for each target.
#[derive(Debug, Clone)]
pub struct CliSettings {
    pub config_path: String,
    pub be_quiet: Option<bool>,
    pub input: Option<String>,
    pub output: Option<String>,
}
/// Runs some options from command line <br>
/// Set TOML config-path to load settings
pub fn parse_args() -> CliSettings {
    let args = Cli::parse();
    if args.version {
        println!("{} version: {}", PKG_NAME, PKG_VERSION);
        std::process::exit(0);
    }
    let config_path = args
        .config_path
        .unwrap_or(String::from(DEFAULT_CONFIG_PATH));
    CliSettings {
        config_path,
        be_quiet: if args.be_quiet { Some(true) } else { None },
        input: args.input_path,
        output: args.output_path,
    }
}
