//! ## Metrics Data for "Consumer Metrics"
//! - Collecting, grouping and differentiate name, values, etc.

use chrono::{DateTime, Utc};
use std::{cmp::Ordering, time::SystemTime};

/// Associate type for metric name in [`MetricsConsumerData`] <br>
/// It allows to collect several **values** related with same **time**
#[derive(Clone, Debug)]
pub struct MetricTimeData {
    pub time: SystemTime,
    pub values: Vec<f64>,
}
/// Magic rust **enum** to clasify metrics names and their associated values in [`MetricTimeData`] <br>
/// - It can combine different items with different types with associated values
/// - Can be extended easily, **rust** will enforce consistence and definitions
#[derive(Clone, Debug, Default)]
pub enum MetricsConsumerData {
    Mem(MetricTimeData),
    Cpu(MetricTimeData),
    #[default]
    Unknown,
}
/// As **sort** is needed for some use cases, like **output** <br>
/// Some implementations has to be written here, are not auto generated via **derive** macros, associtated types with types like **f64** will not allow autogeneration
impl Eq for MetricsConsumerData {}
impl Ord for MetricsConsumerData {
    fn cmp(&self, other: &Self) -> Ordering {
        self.name().cmp(&other.name())
    }
}
impl PartialOrd for MetricsConsumerData {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.name().cmp(&other.name()))
    }
}
impl PartialEq for MetricsConsumerData {
    fn eq(&self, other: &Self) -> bool {
        self.name() == other.name()
    }
}
/// Display per item here
impl std::fmt::Display for MetricsConsumerData {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            MetricsConsumerData::Mem(data) => write!(
                f,
                "{} mem {}",
                DateTime::<Utc>::from(data.time).format("%Y-%m-%dT%H:%M:%SZ"),
                format!("{:?}", data.values)
                    .replace("[", "")
                    .replace("]", "")
            ),
            MetricsConsumerData::Cpu(data) => write!(
                f,
                "{} cpu {}",
                DateTime::<Utc>::from(data.time).format("%Y-%m-%dT%H:%M:%SZ"),
                format!("{:?}", data.values)
                    .replace("[", "")
                    .replace("]", "")
            ),
            MetricsConsumerData::Unknown => write!(f, "anonymous"),
        }
    }
}
/// Some implementations:
/// - to get item **name** or **data** values
/// - to **add_data** to exiting [`MetricTimeData`] **values**
/// - **from_values** allows create an **enum** item
impl MetricsConsumerData {
    pub fn name(&self) -> String {
        match self {
            MetricsConsumerData::Mem(_) => String::from("mem"),
            MetricsConsumerData::Cpu(_) => String::from("cpu"),
            MetricsConsumerData::Unknown => String::from("unknown"),
        }
    }
    pub fn time_data(&self) -> Option<MetricTimeData> {
        match self {
            MetricsConsumerData::Mem(data) => Some(data.to_owned()),
            MetricsConsumerData::Cpu(data) => Some(data.to_owned()),
            MetricsConsumerData::Unknown => None,
        }
    }
    pub fn add_data(&mut self, value: f64) {
        match self {
            MetricsConsumerData::Mem(data) => data.values.push(value),
            MetricsConsumerData::Cpu(data) => data.values.push(value),
            MetricsConsumerData::Unknown => (),
        }
    }
    pub fn from_values(name: &str, time: SystemTime, values: Vec<f64>) -> MetricsConsumerData {
        let metric_time_data = MetricTimeData { time, values };
        match name {
            "mem" | "Mem" | "MEM" => MetricsConsumerData::Mem(metric_time_data),
            "cpu" | "Cpu" | "CPU" => MetricsConsumerData::Cpu(metric_time_data),
            "unknown" | "Unknown" => MetricsConsumerData::Unknown,
            _ => MetricsConsumerData::default(),
        }
    }
}
