//! # Backend internal interview (rust) - improved
//!
//! This **Improved** branch is a rather disruptive approach to the [initial proposal](https://repo.jesusperez.pro/NewRelic/be-technical-interview-rust)<br>
//! [Branch basic-fixed](https://repo.jesusperez.pro/NewRelic/be-technical-interview-rust/src/branch/basic-fixed) tried to solve proposal from initial code
//! as a **continuity effort** with the necessary changes and some improvement adjustments such as the **parallel input processing**
//!
//! ## In summary
//! - [x]  Define a basic model, easily to extend and modify. **Abstraction / Generic**.
//! - [x]  Structs and implementations to specific metricis traitments. **Modular appoach**.
//! - [x]  Settings and configuration for interactive and non interactive processing (batch mode) **Customize on context**.
//!
//! ## Main improvements
//! - Create abstractions / models to load and parse metrics, can be easily extended
//! - Define basic operations via [`MetricParser`] trait path **defs/metrics.rs**
//! - Use **structs** to group metrics attributes and implement operations like [`metrics_consumer::MetricsConsumerTarget`]
//! - Use **enums** with associated values like [`defs::metric_data::MetricsConsumerData`] path  **defs/metric_data** to group attributes and values
//! - Remove **Maps collections**, use [Vectors](https://doc.rust-lang.org/std/vec/struct.Vec.html)
//! - Use **const** for DEFAULT values on [`defs`] module
//! - Remove all <i>hardcoded</i> values to [`Config`] settings, a **TOML config file** (see config.toml) can be used to define several metrics input and customize output
//! - **Output** write to file (append o rewrite mode) is working (for now I/O is sync)
//! - **Command line** arguments are processed [`parse_args`] with [`defs::CliSettings`] via [clap](https://docs.rs/clap/latest/clap/) and **overide** the ones loaded in **config files**
//! -  **Tests** have been accommodated from previous [Branch basic-fixed](https://repo.jesusperez.pro/NewRelic/be-technical-interview-rust/src/branch/basic-fixed) version to **imporved** approach
//!
//! ## Benefits
//!
//! | Element          | Benefit |
//! |----------------- |---------|
//! | Generic traits   | Group generic task for metric processing, steps / tasks separation |
//! | Structs          | Customize atributes and implementation for specific target or patterns  |
//! | Enums with values| Associate attributes metrics and values, easy to add new attributes or combine with different values at once |
//! | Vectors          | Simplify types / grouped in structs, priorize vectors type, easy to iterate, filter, sort, etc |
//! | Const and Config | Group main const, define metric targes and operations in declarative mode for non intective traitment |
//! | Command line args| Help to run in terminal as a cli |
//! | Unit Tests       | Verify some operations results |
//!
//! ## Ideas not included
//! - **Async I/O** to scale and performance ?
//! - Other Thread alternatives like [Tokio](https://tokio.rs/) or/and [Coroutines](https://doc.rust-lang.org/std/ops/trait.Coroutine.html)
//! - Benchmarking for optimization
//! - More **tests**
//! - Run as **API mode** not only as batch processing
//! <br>
//!
//! Code is in a private repository with several branches[^note].
//! [^note]: Link to [branch repository improved](https://repo.jesusperez.pro/NewRelic/be-technical-interview-rust/src/branch/improved)

#![doc(html_logo_url = "https://info.jesusperez.pro/img/jesusperez-logo-b.png")]
#![doc = include_str!("../assets/howto.md")]
use crate::defs::{
    load_from_file, parse_args, Config, MetricParser, CFG_FILE_EXTENSION, DEFAULT_CONFIG_PATH,
    DEFAULT_INPUT_PATH, DEFAULT_REG_EXP, PKG_NAME, PKG_VERSION,
};
use crate::metrics_consumer::MetricsConsumerTarget;
#[doc = include_str!("../README.md")]
use std::{
    sync::{
        mpsc,
        mpsc::{Receiver, Sender},
    },
    thread::spawn,
    time::Instant,
};

mod defs;
mod metrics_consumer;

// Tests are in a separated module for easy access
#[cfg(test)]
mod tests;

/// Main threads control flow for each `target_list` item <br>
/// All process are collected and finally <br>
/// [`MetricsConsumerTarget`] as **T** to **show_metrics** is called to get **print** or **write** results <br>
/// **be_quiet** attribute is just to avoid (true) all messages around processing and parsing operations <br>
/// ## For paralellism
/// This can be done with [Tokio](https://tokio.rs/) as alternative to [std::thread](https://doc.rust-lang.org/std/thread/) <br>
/// It will require load other **crates** and feature customizations<br>
/// Another alternative could be [Coroutines](https://doc.rust-lang.org/std/ops/trait.Coroutine.html) <br>
/// As experimental features is a **nightly-only** (October 2024)
fn generate_metrics<T: MetricParser + Sync + Send + 'static>(targets_list: Vec<T>, be_quiet: bool) {
    let n_items = targets_list.len();
    let mut input_threads = Vec::with_capacity(n_items);
    let (tx, rx): (Sender<Option<T>>, Receiver<Option<T>>) = mpsc::channel();
    targets_list.into_iter().for_each(|metrics_item| {
        let thread_tx = tx.clone();
        input_threads.push(spawn(move || {
            let start = Instant::now();
            match metrics_item.load_input() {
                Ok(mut result) => {
                    result.collect_aggregates();
                    thread_tx.send(Some(result)).unwrap_or_default();
                }
                Err(err) => {
                    eprint!("Error: {}", err);
                    thread_tx.send(None).unwrap_or_default();
                }
            }
            if !be_quiet {
                println!(
                    "Processing {} took: {:?} ms",
                    &metrics_item.input(),
                    start.elapsed().as_millis()
                )
            }
        }));
    });
    let mut inputs_metrics = Vec::with_capacity(n_items);
    for _ in 0..input_threads.len() {
        match rx.recv() {
            Ok(result) => inputs_metrics.push(result),
            Err(e) => eprint!("Error: {}", e),
        }
    }
    for thread in input_threads {
        let _ = thread.join();
    }
    inputs_metrics.iter().for_each(|metrics_item| {
        if let Some(metrics_data) = metrics_item {
            if !be_quiet {
                println!("\n{}: ---------------\n", &metrics_data.input());
            }
            let _ = metrics_data.show_metrics();
        }
    });
}
/// Keep it short:
/// - Parse [`defs::CliSettings`] from command-line arguments
/// - Load [`Config`] file settings from **config_path**
/// - Override [`Config`] settings with cli arguments parsed
/// - Create **targets list** as [`MetricsConsumerTarget`] vector
/// - Call to [`generate_metrics`] to do the job
/// - If not **be_quiet** mode print elapsed time in milliseconds
///
/// > This is not running async and not expect any **Result**.
fn main() {
    let main_start = Instant::now();
    let args_settings = parse_args();
    let config: Config = if std::path::Path::new(&args_settings.config_path).exists() {
        load_from_file(&args_settings.config_path).unwrap_or_else(|e| {
            eprintln!("Settings error: {}", e);
            Config::default()
        })
    } else {
        Config::default()
    };
    let config = config.add_cli_settings(args_settings.clone());
    if !config.be_quiet {
        println!("Loaded config from: {}", &args_settings.config_path);
    }
    let targets_list = config
        .targets
        .iter()
        .map(|item| MetricsConsumerTarget {
            input: String::from(&item.input),
            output: String::from(&item.output),
            out_overwrite: item.out_overwrite,
            reg_exp: String::from(&item.reg_exp),
            ..Default::default()
        })
        .collect();
    generate_metrics(targets_list, config.be_quiet);
    if !config.be_quiet {
        println!(
            "\nALL Processing took: {:?} ms",
            main_start.elapsed().as_millis()
        )
    }
}
