//! ## MetricsConsumerTarget definitions and implementations
//! Specific metric class **Consumer Metric** using generic metric operations <br>
//! Save source / result paths, parsing regular_expression and collect metrics values and their aggreates using [`MetricsConsumerData`]
//! From this modular approach other metrics classes can be defined and implemented

use regex::Regex;
use std::{
    error::Error,
    fs::File,
    io::{prelude::*, BufReader, BufWriter},
    time::{Duration, UNIX_EPOCH},
};

use crate::defs::{MetricParser, MetricsConsumerData};

/// Attributes definition
/// - **reg_exp** has some difficulties to be used with [regex](https://docs.rs/regex/latest/regex/), it works better with [string literals](https://doc.rust-lang.org/reference/expressions/literal-expr.html#string-literal-expressions)
/// - **metrics** and **aggregates** are vectors of [`MetricsConsumerData`] enums values to group input lines or save aggregates values
/// - **metrics** and **aggregates** have same type for implementation simplification, **aggregates** only use <u>first vector value</u> <br>
///   it can be easily used or extended to also save other computed values like: max, min, etc.
#[derive(Debug, Clone, Default)]
pub(crate) struct MetricsConsumerTarget {
    pub input: String,
    pub output: String,
    pub out_overwrite: bool,
    pub reg_exp: String,
    pub metrics: Vec<MetricsConsumerData>,
    pub aggregates: Vec<MetricsConsumerData>,
}

/// Implement generic metrics operations / tasks for [`MetricsConsumerData`]
///
impl MetricParser for MetricsConsumerTarget {
    fn parse(&self, file: File, reg_exp: Regex) -> Self {
        let buf = BufReader::new(file);
        let mut consumer_target = self.clone();
        buf.lines()
            .enumerate()
            .for_each(|(index, read_line)| match read_line {
                Ok(line) => {
                    if let Some(metric_line) = self.parse_line(&line, index, &reg_exp) {
                        let (timestamp, name, value) = metric_line;
                        let minute =
                            UNIX_EPOCH + Duration::from_secs((timestamp - (timestamp % 60)) as u64);
                        let mut not_found = true;
                        for metric in consumer_target.metrics.iter_mut() {
                            if metric.name() == name.to_lowercase() {
                                if let Some(metric_data) = metric.time_data() {
                                    if metric_data.time == minute {
                                        metric.add_data(value);
                                        not_found = false;
                                        break;
                                    }
                                }
                            }
                        }
                        if not_found {
                            consumer_target
                                .metrics
                                .push(MetricsConsumerData::from_values(
                                    &name,
                                    minute,
                                    Vec::from([value]),
                                ));
                        }
                    }
                }
                Err(e) => {
                    eprintln!("Error reading line {}: {}", index, e);
                }
            });
        consumer_target
    }
    fn load_input(&self) -> Result<Self, Box<dyn Error>> {
        let file = File::open(&self.input)
            .map_err(|err| format!("Error reading file: {} {}", &self.input, err))?;
        if self.reg_exp.is_empty() {
            return Err(String::from("Error invalid reg expression").into());
        }
        let reg_exp = Regex::new(&self.reg_exp)?;

        Ok(self.parse(file, reg_exp))
    }
    fn collect_aggregates(&mut self) {
        self.metrics.iter().for_each(|metric_data| {
            let name = metric_data.name();
            if let Some(metric_time_data) = metric_data.time_data() {
                let average = metric_time_data.values.iter().sum::<f64>()
                    / metric_time_data.values.len() as f64;
                let mut not_found = true;
                for metric in self.aggregates.iter_mut() {
                    if metric.name() == name.to_lowercase() {
                        if let Some(metric_data) = metric.time_data() {
                            if metric_data.time == metric_time_data.time {
                                metric.add_data(average);
                                not_found = false;
                                break;
                            }
                        }
                    }
                }
                if not_found {
                    self.aggregates.push(MetricsConsumerData::from_values(
                        &name,
                        metric_time_data.time,
                        Vec::from([average]),
                    ));
                }
            }
        })
    }
    fn show_metrics(&self) -> Result<Vec<String>, std::io::Error> {
        let mut output = Vec::new();
        let mut aggregates = self.aggregates.clone();
        aggregates.sort();
        aggregates.iter().for_each(|metric_data| {
            if metric_data.time_data().is_some() {
                let output_line = format!("{}", metric_data);
                match self.output.as_str() {
                    "vec" => output.push(output_line),
                    "print" | "" => println!("{}", output_line),
                    _ => output.push(output_line),
                }
            }
        });
        match self.output.as_str() {
            "vec" | "print" | "" => return Ok(output.to_owned()),
            _ => {
                if self.out_overwrite && std::path::Path::new(&self.output).exists() {
                    std::fs::remove_file(&self.output)?;
                }
                if !std::path::Path::new(&self.output).exists() {
                    File::create(&self.output)?;
                }
                let file = File::options().append(true).open(&self.output)?;
                let mut writer = BufWriter::new(file);
                writer.write_all(output.join("\n").as_bytes())?;
                let text_overwrite = if self.out_overwrite {
                    String::from("overwriten")
                } else {
                    String::from("")
                };
                println!(
                    "Metrics for '{}' are saved in '{}' {}",
                    &self.input, &self.output, &text_overwrite
                );
            }
        };
        Ok(output.to_owned())
    }
    fn input(&self) -> String {
        self.input.to_string()
    }
}
/// Specific implementations like **parse_line**
impl MetricsConsumerTarget {
    fn show_invalid_line(&self, index: usize, line: &str) {
        println!("invalid line: {} {}", index, line);
    }
    /// Check metric line values
    pub fn parse_line(&self, line: &str, index: usize, re: &Regex) -> Option<(i64, String, f64)> {
        if let Some(caps) = re.captures(line) {
            let timestamp = match caps[1].parse::<i64>() {
                Ok(value) => value,
                Err(e) => {
                    println!("Parse timestamp {} error {}", &caps[1], e);
                    self.show_invalid_line(index, line);
                    return None;
                }
            };
            let metric_value = match caps[3].parse::<f64>() {
                Ok(value) => value,
                Err(e) => {
                    println!("Parse metric_value {} error {}", &caps[3], e);
                    self.show_invalid_line(index, line);
                    return None;
                }
            };
            Some((timestamp, caps[2].to_string(), metric_value))
        } else {
            self.show_invalid_line(index, line);
            None
        }
    }
}
